# -*- coding: utf-8 -*-
from easy_karabiner import util
from easy_karabiner import lookup
from easy_karabiner.xml_base import XML_base
from easy_karabiner.def_filter_map import get_name_tag_by_def_tag


class BaseDef(XML_base):
    def __init__(self, name, *tag_vals, **kwargs):
        self.name = name
        self.tag_vals = tag_vals
        self.kwargs = kwargs

    def get_name(self):
        return self.name

    def get_def_tag_name(self):
        return '%sdef' % self.__class__.__name__.lower()

    def get_name_tag_name(self):
        return get_name_tag_by_def_tag(self.get_def_tag_name())

    def get_tag_val_pair(self, val):
        raise Exception("Need override")

    def get_tag_val_pairs(self, tag_vals):
        return list(map(self.get_tag_val_pair, tag_vals))

    def split_name_and_attrs(self, tag_name):
        name_parts = tag_name.split()

        if len(name_parts) > 1:
            # transform `key="value"` to `(key, value)`
            def to_pair(s, del_chars=('"', "'")):
                cleaned = filter(lambda c: c not in del_chars, s)
                return ''.join(cleaned).split('=')

            tag_attrs = dict(map(to_pair, name_parts[1:]))
        else:
            tag_attrs = {}

        name = name_parts[0]
        return (name, tag_attrs)

    def to_xml(self):
        xml_tree = self.create_tag(self.get_def_tag_name())
        name_tag = self.create_tag(self.get_name_tag_name(), self.name)
        xml_tree.append(name_tag)

        tag_val_pairs = self.get_tag_val_pairs(self.tag_vals)

        for tag_name, tag_val in tag_val_pairs:
            if len(tag_name) > 0:
                tag_name, tag_attrs = self.split_name_and_attrs(tag_name)
                tag = self.create_tag(tag_name, tag_val, attrib=tag_attrs)
                xml_tree.append(tag)

        return xml_tree


class App(BaseDef):
    """
    >>> d = App('BILIBILI', 'com.typcn.Bilibili')
    >>> s = '''
    ...    <appdef>
    ...      <appname>BILIBILI</appname>
    ...      <equal>com.typcn.Bilibili</equal>
    ...    </appdef>'''
    >>> util.assert_xml_equal(d, s)
    """

    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        if val[0] == '.':
            return ('suffix', val)
        elif val[-1] == '.':
            return ('prefix', val)
        else:
            return ('equal', val)


class WindowName(BaseDef):
    def get_tag_val_pair(self, val):
        return ('regex', val)


class DeviceVendor(BaseDef):
    def get_tag_val_pair(self, val):
        return ('vendorid', val)


class DeviceProduct(BaseDef):
    def get_tag_val_pair(self, val):
        return ('productid', val)


class InputSource(BaseDef):
    def is_host(self, val):
        parts = val.split('.')
        return len(parts) >= 3 and all(map(len, parts))

    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        if val[-1] == '.':
            return ('inputsourceid_prefix', val[:-1])
        elif self.is_host(val):
            return ('inputsourceid_equal', val)
        else:
            return ('languagecode', val)


class VKChangeInputSource(InputSource):
    pass


class VKOpenURL(BaseDef):
    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        elif val[0] == '/':
            return ('url type="file"', val)
        elif val.startswith('#!'):
            if val.startswith('#! '):
                val = val[3:]
            script = self.create_cdata_text(val)
            return ('url type="shell"', script)
        else:
            return ('url', val)

    def to_xml(self):
        xml_tree = super(VKOpenURL, self).to_xml()
        if self.kwargs.get('background', False):
            background_tag = self.create_tag('background')
            xml_tree.append(background_tag)
        return xml_tree


class Replacement(BaseDef):
    def get_tag_val_pair(self, val):
        return ('replacementvalue', val)

    def get_tag_val_pairs(self, tag_vals):
        return [self.get_tag_val_pair(', '.join(tag_vals))]


class UIElementRole(BaseDef):
    def to_xml(self):
        xml_tree = self.create_tag(self.get_def_tag_name(), self.name)
        return xml_tree


_CLASSES = util.find_all_subclass_of(BaseDef, globals())

def is_vkopenurl_format(name):
    parts = name.split(':', 2)
    return parts[0].lower() == 'open' or name.startswith('KeyCode::VK_OPEN_URL_')

def get_vkopenurl_defname(name):
    # VKOpenURL defname must start with 'KeyCode::VK_OPEN_URL_'
    if name.startswith('KeyCode::VK_OPEN_URL_'):
        defname = name
    else:
        defname = 'KeyCode::VK_OPEN_URL_%s' % name.split(':', 2)[-1]

    return defname

def is_replacement_format(name):
    return name.startswith('{{') and name.endswith('}}')

def get_replacement_defname(name):
    return name[2:-2]

def split_clsname_defname(name):
    parts = name.split(':', 2)

    # Open::defname
    # KeyCode::VK_OPEN_URL_defname
    if is_vkopenurl_format(name):
        clsname = 'VKOpenURL'
        defname = get_vkopenurl_defname(name)
    # {{defname}}
    elif is_replacement_format(name):
        clsname = 'Replacement'
        defname = get_replacement_defname(name)
    # clsname::defname
    elif len(parts) > 2:
        clsname = lookup.get_alias('DEF_ALIAS', parts[0], parts[0])
        defname = parts[-1]
    # defname
    else:
        clsname = 'App'
        defname = name

    return (clsname, defname)

def get_ground_truth_vals(clsname, vals):
    if clsname == 'VKOpenURL':
        def get_apppath(v):
            return v if v.startswith('/') else util.get_apppath(v, v)

        vals = list(map(get_apppath, vals))
    return vals

def parse_definition(name, vals):
    definition = None
    clsname, defname = split_clsname_defname(name)
    vals = get_ground_truth_vals(clsname, vals)

    for cls in _CLASSES:
        if clsname == cls.__name__:
            definition = cls(defname, *vals)
            break

    if definition is None:
        raise Exception('Unsupport definition type')

    lookup.DefQuery.add_def(definition)
    return definition


if __name__ == "__main__":
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
"""A tool to generate key remap configuration for Karabiner

Usage:
    easy_karabiner [-evr] [SOURCE] [TARGET | --string]
    easy_karabiner [--help | --version]
"""
from __future__ import print_function
import os
import click
from hashlib import sha1
from subprocess import call
from easy_karabiner import lookup
from easy_karabiner import __version__
from easy_karabiner.xml_base import XML_base
from easy_karabiner.generator import Generator


DEFAULT_CONFIG_PATH = '~/.easy_karabiner.py'
DEFAULT_OUTPUT_PATH = '~/Library/Application Support/Karabiner/private.xml'

DEFAULT_CONFIG_PATH = os.path.expanduser(DEFAULT_CONFIG_PATH)
DEFAULT_OUTPUT_PATH = os.path.expanduser(DEFAULT_OUTPUT_PATH)


@click.command()
@click.help_option('--help', '-h')
@click.version_option(__version__, '--version', '-v', message='%(version)s')
@click.argument('inpath', default=DEFAULT_CONFIG_PATH, type=click.Path())
@click.argument('outpath', default=DEFAULT_OUTPUT_PATH, type=click.Path())
@click.option('--verbose', '-V', help='Print more text.', is_flag=True)
@click.option('--string', '-s', help='Output as string.', is_flag=True)
@click.option('--reload', '-r', help='Reload Karabiner.', is_flag=True)
@click.option('--edit', '-e', help='Edit default config file.', is_flag=True)
def main(inpath, outpath, **options):
    """
    \b
    $ easy_karabiner
    $ easy_karabiner input.py output.xml
    $ easy_karabiner input.py --string
    """
    global VERBOSE
    VERBOSE = options.get('verbose')

    if options.get('edit'):
        edit_config_file()
        exit(0)

    try:
        configs = read_config_file(inpath)
        xml_str = gen_config(configs)

        if options.get('string'):
            print(xml_str)
            exit(0)

        try:
            if not is_generated_by_easy_karabiner(outpath):
                backup_file(outpath)
        except IOError:
            pass

        write_generated_xml(outpath, xml_str)

        if options.get('reload') or (outpath == DEFAULT_OUTPUT_PATH):
            reload_karabiner()

        exit(0)
    except IOError:
        print("%s not exist" % inpath)
        exit(1)


def read_config_file(config_path):
    configs = {}
    with open(config_path, 'rb') as fp:
        if VERBOSE:
            print("Execute %s" % config_path)
        exec(compile(fp.read(), config_path, 'exec'), {}, configs)
    return configs

def write_generated_xml(outpath, content):
    with open(outpath, 'wb') as fp:
        if VERBOSE:
            print("Write generated XML config to %s" % outpath)
        fp.write(content)

def edit_config_file():
    editor = os.environ.get('EDITOR', 'vi')
    call([editor, DEFAULT_CONFIG_PATH])

def reload_karabiner():
    if VERBOSE:
        print("Reload Karabiner config")
    call(['karabiner', 'enable', 'private.easy_karabiner'])
    call(['karabiner', 'reloadxml'])

def update_aliases(configs):
    if VERBOSE:
        print("Update alias database")
    # user may define it's own alias table
    alias_names = filter(lambda k: k.endswith('_ALIAS'), configs.keys())
    for alias_name in alias_names:
        lookup.update_alias(configs[alias_name])

def gen_config(configs):
    remaps = configs.get('REMAPS', [])
    definitions = configs.get('DEFINITIONS', [])
    update_aliases(configs)
    if VERBOSE:
        print("Generate XML configuration string")
    return Generator(remaps=remaps, definitions=definitions).generate()

def is_generated_by_easy_karabiner(filepath):
    tag = XML_base.parse(filepath).find('Easy-Karabiner')
    return tag is not None

def backup_file(filepath):
    with open(filepath, 'rb') as fp:
        checksum = sha1(fp.read()).hexdigest()[:7]
        parts = os.path.basename(filepath).split('.')
        parts.insert(-1, checksum)
        newname = os.path.join(os.path.dirname(filepath), '.'.join(parts))
        if VERBOSE:
            print("Backup original XML config file")
        # private.xml -> private.941f123.xml
        os.rename(filepath, newname)


# -*- coding: utf-8 -*-
from __future__ import print_function
from easy_karabiner import util
from easy_karabiner import lookup
from easy_karabiner.xml_base import XML_base
from easy_karabiner.key import Key


class BaseKeyToKey(XML_base, util.Hashable):
    MULTI_KEYS_FMT = '@begin\n{key}\n@end'
    AUTOGEN_FMT = ' {type}\n{keys_str}\n'

    def __init__(self, *keys, **kwargs):
        self.keys_str = self.parse(*keys, **kwargs)

    def parse(self, *keys, **kwargs):
        raise Exception("Need override")

    def get_type(self):
        return '__%s__' % self.__class__.__name__

    def to_xml(self):
        text = self.AUTOGEN_FMT.format(type=self.get_type(),
                                       keys_str=self.keys_str)
        return self.create_tag('autogen', text)

    def _id(self):
        return (self.get_type(), self.keys_str)

class UniversalKeyToKey(BaseKeyToKey):
    def __init__(self, type, *keys, **kwargs):
        self.type = type
        super(UniversalKeyToKey, self).__init__(*keys, **kwargs)

    def get_type(self):
        return self.type

    def parse(self, *keys, **kwargs):
        toKey = lambda k: Key(k, keep_first_keycode=True, has_modifier_none=False)
        keys = map(toKey, keys)
        keys = map(str, keys)

        return ',\n'.join(keys)


class _KeyToMultiKeys(BaseKeyToKey):
    KEYS_FMT = ('{from_key},\n'
                '@begin\n'
                '{to_key}\n'
                '@end\n'
                '@begin\n'
                '{additional_key}\n'
                '@end')

    def parse(self, from_key, additional_key, to_key=None, has_modifier_none=True):
        if to_key:
            to_key = Key(to_key)
        else:
            to_key = Key(from_key)

        from_key = Key(from_key, has_modifier_none)
        additional_key = Key(additional_key)

        return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                    to_key=to_key.to_str(),
                                    additional_key=additional_key.to_str())

class _OneKeyEvent(BaseKeyToKey):
    def parse(self, key, has_modifier_none=False):
        return Key(key, has_modifier_none).to_str()

class _ZeroKeyEvent(BaseKeyToKey):
    def parse(self):
        return ''

class _MultiKeys(UniversalKeyToKey):
    def __init__(self, *keys, **kwargs):
        super(_MultiKeys, self).__init__(self.get_type(), *keys, **kwargs)

    def get_type(self):
        return '__%s__' % self.__class__.__name__


class KeyToKey(BaseKeyToKey):
    '''
    >>> print(KeyToKey('alt', 'cmd'))
    <autogen>__KeyToKey__ KeyCode::OPTION_L, KeyCode::COMMAND_L</autogen>
    '''

    KEYS_FMT = '{from_key},\n{to_key}'

    def parse(self, from_key, to_key, has_modifier_none=True):
        from_key = Key(from_key, has_modifier_none=has_modifier_none)
        to_key = Key(to_key)

        return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                    to_key=to_key.to_str())

class DropAllKeys(BaseKeyToKey):
    KEYS_FMT = '{from_key},\n{options}'

    def parse(self, from_modifier, options=None):
        from_key = Key(from_modifier, keep_first_keycode=True)

        if options is None:
            return from_key.to_str()
        else:
            options = Key(options, keep_first_keycode=True)
            return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                        options=options.to_str())

class SimultaneousKeyPresses(BaseKeyToKey):
    KEYS_FMT = ('@begin\n'
                '{from_key}\n'
                '@end\n'
                '@begin\n'
                '{to_key}\n'
                '@end')

    def parse(self, from_key, to_key):
        from_key = Key(from_key, False)
        to_key = Key(to_key, False)

        return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                    to_key=to_key.to_str())


class DoublePressModifier(_KeyToMultiKeys):
    pass

class HoldingKeyToKey(_KeyToMultiKeys):
    pass

class KeyOverlaidModifier(_KeyToMultiKeys):
    pass

class KeyDownUpToKey(_KeyToMultiKeys):
    def parse(self, from_key, immediately_key, interrupted_key):
        return super(KeyDownUpToKey, self).parse(from_key,
                                                 interrupted_key,
                                                 immediately_key)


class BlockUntilKeyUp(_OneKeyEvent):
    pass

class DropKeyAfterRemap(_OneKeyEvent):
    pass


class PassThrough(_ZeroKeyEvent):
    pass

class DropPointingRelativeCursorMove(_ZeroKeyEvent):
    pass


# NOTICE: because below `autogen` command not yet documented,
#         so those implements maybe changed in future
class DropScrollWheel(_MultiKeys):
    pass

class FlipPointingRelative(_MultiKeys):
    pass

class FlipScrollWheel(_MultiKeys):
    pass

class IgnoreMultipleSameKeyPress(_MultiKeys):
    pass

class StripModifierFromScrollWheel(_MultiKeys):
    pass

class ScrollWheelToScrollWheel(_MultiKeys):
    pass

class ScrollWheelToKey(_MultiKeys):
    pass

class PointingRelativeToScroll(_MultiKeys):
    pass

class PointingRelativeToKey(_MultiKeys):
    pass

class ForceNumLockOn(_MultiKeys):
    pass

class ShowStatusMessage(BaseKeyToKey):
    def parse(self, *keys, **kwargs):
        return ' '.join(keys)

class SetKeyboardType(_OneKeyEvent):
    pass


_CLASSES = util.find_all_subclass_of(BaseKeyToKey, globals())

def get_ground_truth_vals(clsname, vals):
    def replace_vkopenurl_start(val):
        return util.replace_startswith_to(val, 'Open::', 'KeyCode::VK_OPEN_URL_')

    vals = list(map(replace_vkopenurl_start, vals))
    return vals


def parse_keymap(command, keys):
    clsname = lookup.get_alias('KEYMAP_ALIAS', command, command.strip('_'))
    keys = get_ground_truth_vals(clsname, keys)

    for cls in _CLASSES:
        if clsname == cls.__name__:
            return cls(*keys)
    return UniversalKeyToKey(command, *keys)


if __name__ == "__main__":
    import doctest
    doctest.testmod()


DEF_ALIAS = {
   'window':  'WindowName',
   'open': 'VKOpenURL',
   'inputsource': 'ChangeInputSource',
}

KEYMAP_ALIAS = {
   'double':  'DoublePressModifier',
   'holding': 'HoldingKeyToKey',
   'press_modifier': 'KeyOverlaidModifier',
}

MODIFIER_ALIAS = {
    "shift"   : "SHIFT_L",
    "shift_l" : "SHIFT_L",
    "shift_r" : "SHIFT_R",
    "cmd"     : "COMMAND_L",
    "command" : "COMMAND_L",
    "cmd_l"   : "COMMAND_L",
    "cmd_r"   : "COMMAND_R",
    "opt"     : "OPTION_L",
    "option"  : "OPTION_L",
    "opt_l"   : "OPTION_L",
    "opt_r"   : "OPTION_R",
    "alt"     : "OPTION_L",
    "alt_l"   : "OPTION_L",
    "alt_r"   : "OPTION_R",
    "ctrl"    : "CONTROL_L",
    "control" : "CONTROL_L",
    "ctrl_l"  : "CONTROL_L",
    "ctrl_r"  : "CONTROL_R",
}

KEY_ALIAS = {
    "whitespace" : "SPACE",
    "sp"         : "SPACE",
    "del"        : "DELETE",
    "fdel"       : "FORWARD_DELETE",
    "esc"        : "ESCAPE",
    "caps"       : "CAPSLOCK",
    "left"       : "CURSOR_LEFT",
    "right"      : "CURSOR_RIGHT",
    "down"       : "CURSOR_DOWN",
    "up"         : "CURSOR_UP",
    "`"          : "BACKQUOTE",
    "1"          : "KEY_1",
    "2"          : "KEY_2",
    "3"          : "KEY_3",
    "4"          : "KEY_4",
    "5"          : "KEY_5",
    "6"          : "KEY_6",
    "7"          : "KEY_7",
    "8"          : "KEY_8",
    "9"          : "KEY_9",
    "0"          : "KEY_0",
    "-"          : "MINUS",
    "="          : "EQUAL",
    "["          : "BRACKET_LEFT",
    "]"          : "BRACKET_RIGHT",
    "\\"         : "BACKSLASH",
    ";"          : "SEMICOLON",
    "'"          : "QUOTE",
    ","          : "COMMA",
    "."          : "DOT",
    "/"          : "SLASH",

    "mouse_left"   : "PointingButton::LEFT",
    "mouse_right"  : "PointingButton::RIGHT",
    "mouse_middle" : "PointingButton::MIDDLE",
    "mouse1"       : "PointingButton::LEFT",
    "mouse2"       : "PointingButton::RIGHT",
    "mouse3"       : "PointingButton::MIDDLE",

    "scroll_up"    : "ScrollWheel::UP",
    "scroll_down"  : "ScrollWheel::DOWN",
    "scroll_left"  : "ScrollWheel::left",
    "scroll_right" : "ScrollWheel::right",
}

KEY_ALIAS.update(MODIFIER_ALIAS)


# -*- coding: utf-8 -*-

_DEF_FILTER_MAP = {
    'appdef'                 : 'Filter',
    'replacementdef'         : 'ReplacementFilter',
    'windownamedef'          : 'WindowNameFilter',
    'devicevendordef'        : 'DeviceVendorFilter',
    'deviceproductdef'       : 'DeviceProductFilter',
    'inputsourcedef'         : 'InputSourceFilter',
    'vkchangeinputsourcedef' : 'InputSourceFilter',
}

_DEFTAG_NAMETAG_MAP = {
    'appdef'                 : 'appname',
    'replacementdef'         : 'replacementname',
    'devicevendordef'        : 'vendorname',
    'deviceproductdef'       : 'productname',
    'uielementroledef'       : '',
    'windownamedef'          : 'name',
    'vkopenurldef'           : 'name',
    'inputsourcedef'         : 'name',
    'vkchangeinputsourcedef' : 'name',
}

def get_name_tag_by_def_tag(defname, d=None):
    return _DEFTAG_NAMETAG_MAP.get(defname, d)

def get_filter_by_def(defname, d=None):
    return _DEF_FILTER_MAP.get(defname, d)


# -*- coding: utf-8 -*-
from __future__ import print_function
from easy_karabiner import lookup

class Key(object):
    """Convert space seperated string to Karabiner's favorite format

    >>> print(Key('shift_l Cmd'))
    KeyCode::SHIFT_L, ModifierFlag::COMMAND_L
    """

    def __init__(self, keys='', has_modifier_none=False, keep_first_keycode=False):
        self.has_modifier_none = has_modifier_none
        self.keep_first_keycode = keep_first_keycode
        self.keys = self.parse(keys)

    def parse(self, keys):
        # remove multiple whitespaces and convert to raw value
        keys = " ".join(keys.split()).split()
        # if no alias find, then return the original value
        keys = map(lambda k: lookup.get_alias('KEY_ALIAS', k, k), keys)
        # if no key header find, then return the original value
        keys = map(self.add_key_header, keys)
        # we need adjust position of keys, because
        # the first key can't be modifier key in most case
        keys = self.rearrange_keys(list(keys))

        if len(keys) > 0:
            if not self.keep_first_keycode and self.is_modifier_key(keys[0]):
                # change the header of first key to `KeyCode`
                keys = self.regularize_first_key(keys)
            if self.has_modifier_none and self.is_modifier_key(keys[-1]):
                keys = self.add_modifier_none(keys)

        return keys

    def add_key_header(self, key):
        header = lookup.KeyCodeQuery.query(key.upper())
        if header is None:
            return key
        else:
            return "%s::%s" % (header, key.upper())

    def rearrange_keys(self, keys):
        tmp = []
        last = 0

        for i in range(len(keys)):
            if not self.is_modifier_key(keys[i]):
                tmp.append(keys[i])
                while last < i:
                    tmp.append(keys[last])
                    last += 1
                last = i + 1

        tmp.extend(keys[last:])
        return tmp

    def is_modifier_key(self, key):
        return key.lower().startswith('modifier')

    def regularize_first_key(self, keys):
        parts = keys[0].split(':')
        keys[0] = 'KeyCode::' + parts[-1]
        return keys

    # if there  need append 'ModifierFlag::NONE' if you want change from this key
    # For more information about 'ModifierFlag::NONE', see https://pqrs.org/osx/karabiner/xml.html.en
    def add_modifier_none(self, keys):
        keys.append('ModifierFlag::NONE')
        return keys

    def to_str(self):
        return ', '.join(self.keys)

    def __add__(self, another):
        res = Key()
        res.keys = self.keys + another.keys
        res.has_modifier_none = self.has_modifier_none or another.has_modifier_none
        res.keep_first_keycode = self.keep_first_keycode or another.keep_first_keycode
        return res

    def __str__(self):
        return self.to_str()


if __name__ == "__main__":
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
from __future__ import print_function
import os
from easy_karabiner import alias
from easy_karabiner.xml_base import XML_base
from easy_karabiner.def_filter_map import get_name_tag_by_def_tag, get_filter_by_def


# alias is case-insensitive
def get_alias(tblname, k, d=None):
    return alias.__dict__[tblname].get(k.lower(), d)

def update_alias(tblname, aliases):
    alias.__dict__.setdefault(tblname, {}).update(aliases)
    if tblname == 'MODIFIER_ALIAS':
        alias.KEY_ALIAS.update(alias.MODIFIER_ALIAS)


class BaseQuery(object):
    # NOTICE: need override
    DATA_DIR = None
    DATA_SUFFIX = None
    QUERY_ORDER = None

    def __init__(self):
        self.data = {}

    @classmethod
    def get_instance(cls):
        if hasattr(cls, '_instance'):
            return cls._instance
        else:
            cls._instance = cls()
            cls._instance.load_data()
            return cls._instance

    @classmethod
    def query(cls, value, default=None):
        for k in cls.QUERY_ORDER:
            if cls.is_in(k, value):
                return k
        return default

    @classmethod
    def is_in(cls, k, value):
        if value in cls.get_instance().get(k, []):
            return k
        else:
            return None

    def get(self, k, default=None):
        return self.data.get(k, default or [])

    def load_data(self):
        for type in self.QUERY_ORDER:
            self.data[type] = set(self.get_data(type))
        return self

    def get_datapath(self, type):
        basename = '%s.%s' % (type, self.DATA_SUFFIX)
        return os.path.join(os.path.dirname(__file__), self.DATA_DIR, basename)

    def get_data(self, type):
        raise Exception("Need override")


# Data file come from https://github.com/tekezo/Karabiner/tree/master/src/bridge/generator/keycode/data
class KeyCodeQuery(BaseQuery):
    ''' Query header of key

    >>> print(KeyCodeQuery.query('A'))
    KeyCode

    >>> print(KeyCodeQuery.query('NOREPEAT'))
    Option
    '''

    DATA_DIR = 'data/keycode'
    DATA_SUFFIX = 'data'

    QUERY_ORDER = ['ModifierFlag',
                   'ConsumerKeyCode',
                   'KeyCode',
                   'Option',
                   'InputSource',
                   'PointingButton',
                   'DeviceProduct',
                   'DeviceVendor',
                   'KeyboardType',]


    def get_data(self, type):
        data = []

        with open(self.get_datapath(type), 'r') as fp:
            lines = fp.readlines()
            need_keep = lambda l: not l.startswith('//') and not l.isspace()
            # remove comment line and whitespace line
            lines = filter(need_keep, lines)
            lines = map(lambda l: l.strip(), lines)
            # keep first part
            data = list(map(lambda l: l.split()[0], lines))

        return data


class UndefinedFilterException(Exception):
    pass


# Data file come from https://github.com/tekezo/Karabiner/tree/master/src/core/server/Resources/
class DefQuery(BaseQuery):
    ''' Query definition type of value

    >>> print(DefQuery.query('EMACS_MODE_IGNORE_APPS'))
    replacementdef

    >>> print(DefQuery.query('VIRTUALMACHINE'))
    appdef

    >>> print(DefQuery.query('KeyCode::VK_OPEN_URL_WEB_google'))
    vkopenurldef

    >>> print(DefQuery.query_filter('EMACS_MODE_IGNORE_APPS'))
    Filter
    '''

    DATA_DIR = 'data/def'
    DATA_SUFFIX = 'xml'

    QUERY_ORDER = ['appdef',
                   'replacementdef',
                   'modifierdef',
                   'devicevendordef',
                   'deviceproductdef',
                   'uielementroledef',
                   'windownamedef',
                   'vkopenurldef',
                   'inputsourcedef',
                   'vkchangeinputsourcedef',]

    def get_data(self, type):
        xml_tree = XML_base.parse(self.get_datapath(type))
        name_val = get_name_tag_by_def_tag(type)

        if name_val == '':
            tags = xml_tree.findall(type)
        else:
            tags = xml_tree.findall('%s/%s' % (type, name_val))

        return list(map(lambda tag: tag.text, tags))

    @classmethod
    def add_def(cls, definition):
        self = cls.get_instance()
        type = definition.get_def_tag_name()
        defname = definition.get_name()
        self.data[type].add(defname)

    @classmethod
    def query_filter(cls, def_val):
        if get_alias('MODIFIER_ALIAS', def_val.lower()):
            def_type = 'modifierdef'
        else:
            def_type = cls.query(def_val)

        if def_type is None:
            raise UndefinedFilterException('Undefined filter `%s`' % def_val)
        else:
            return get_filter_by_def(def_type.lower())


if __name__ == '__main__':
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
import lxml.etree as etree
import xml.dom.minidom as minidom


class XML_base(object):
    @staticmethod
    def is_cdata_text(elem):
        return elem.text.startswith('<![CDATA[') and elem.text.endswith(']]>')

    @staticmethod
    def remove_cdata(elem):
        return elem.text[len('<![CDATA['):-len(']]>')]

    @staticmethod
    def create_cdata_text(text):
        return etree.CDATA(text)

    @staticmethod
    def create_tag(name, text=None, **kwargs):
        et = etree.Element(name, **kwargs)
        et.text = text
        return et

    @staticmethod
    def to_format_str(xml_tree, pretty_text=True):
        indent = "  "
        if pretty_text:
            XML_base.pretty_text(xml_tree, indent=indent)
        xml_string = etree.tostring(xml_tree)
        xml_string = minidom.parseString(xml_string).toprettyxml(indent=indent)
        return xml_string

    @staticmethod
    def parse(filepath):
        return etree.parse(filepath).getroot()

    @staticmethod
    def parse_string(xmlstr):
        return etree.fromstring(xmlstr)

    @staticmethod
    def pretty_text(elem, indent="  ", level=0):
        i = "\n" + level * indent

        if len(elem) == 0:
            if elem.text is not None:
                lines = elem.text.split('\n')
                if len(lines) > 1:
                    if not lines[0].startswith(' '):
                        lines[0] = (i + indent) + lines[0]
                    if lines[-1].strip() == '':
                        lines.pop()
                    elem.text = (i + indent).join(lines) + i
        else:
            for subelem in elem:
                XML_base.pretty_text(subelem, indent, level + 1)

        return elem

    def to_xml(self):
        raise Exception('Need override')

    def to_str(self, pretty_text=True, remove_first_line=False):
        xmlstr = self.to_format_str(self.to_xml(), pretty_text=pretty_text)
        if remove_first_line:
            lines = self.to_str().split('\n')[1:]
            return '\n'.join(lines)
        else:
            return xmlstr

    def __str__(self):
        # remove version tag
        return self.to_str(remove_first_line=True)


# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import OrderedDict
from easy_karabiner import __version__
from easy_karabiner.xml_base import XML_base
from easy_karabiner.define import parse_definition
from easy_karabiner.filter import parse_filter
from easy_karabiner.keymap import parse_keymap
from easy_karabiner.lookup import UndefinedFilterException


class Generator(XML_base):
    """
    >>> from easy_karabiner import util
    >>> s1 = Generator().generate()
    >>> s2 = '''
    ...      <root>
    ...        <Easy-Karabiner>{version}</Easy-Karabiner>
    ...        <item>
    ...          <name>Easy-Karabiner</name>
    ...          <item>
    ...            <name>Enable</name>
    ...            <identifier>private.easy_karabiner</identifier>
    ...          </item>
    ...        </item>
    ...      </root>'''.format(version=__version__)
    >>> util.assert_xml_equal(s1, s2)
    """

    ITEM_IDENTIFIER = 'private.easy_karabiner'

    def __init__(self, remaps=None, definitions=None):
        self.remaps = remaps or []
        self.definitions = definitions or dict()

    def init_xml_tree(self):
        version_tag = XML_base.create_tag('Easy-Karabiner', __version__)
        self.xml_tree = XML_base.create_tag('root')
        item_tag = XML_base.create_tag('item')
        self.xml_tree.append(version_tag)
        self.xml_tree.append(item_tag)
        item_tag.append(XML_base.create_tag('name', 'Easy-Karabiner'))
        return item_tag

    def init_subitem_tag(self, item_tag):
        subitem_tag = XML_base.create_tag('item')
        name_tag = XML_base.create_tag('name', 'Enable')
        identifier_tag = XML_base.create_tag('identifier', self.ITEM_IDENTIFIER)
        item_tag.append(subitem_tag)
        subitem_tag.append(name_tag)
        subitem_tag.append(identifier_tag)
        return subitem_tag

    def generate(self):
        item_tag = self.init_xml_tree()

        definitions = self.parse_definitions()
        for d in definitions:
            item_tag.append(d.to_xml())

        blocks = self.parse_remaps()
        subitem_tag = self.init_subitem_tag(item_tag)
        list(map(lambda block: subitem_tag.append(block.to_xml()), blocks))

        return self.to_str()

    def parse_definitions(self):
        definitions = []

        for name, vals in self.definitions.items():
            if isinstance(vals, str):
                vals = [vals]
            definition = parse_definition(name, vals)
            definitions.append(definition)

        return definitions

    def parse_remaps(self):
        tmp = OrderedDict()
        filters_table = {}

        for remap in self.remaps:
            keymap, filters = self.parse_remap(remap)
            filters = tuple(filters)
            filters_table[filters] = filters
            tmp.setdefault(filters, []).append(keymap)

        blocks = []
        for filters in tmp:
            keymaps = tmp[filters]
            filters = filters_table[filters]
            block = Block(keymaps, filters)
            blocks.append(block)

        return blocks

    # return (KeyToKey, [Filter])
    def parse_remap(self, remap):
        if not self.is_list_or_tuple(remap) or len(remap) == 0:
            raise Exception('Wrong format')

        if self.is_list_or_tuple(remap[-1]):
            keyargs = remap[:-1]
            try:
                filters = parse_filter(remap[-1])
            except UndefinedFilterException as e:
                print(e)
                exit(1)
        else:
            keyargs = remap
            filters = []

        # I need check if there exist a command item in here
        # otherwise I cannot judge where is the start position of the `keys` args
        if self.is_command_format(keyargs[0]):
            command = self.get_command_value(keyargs.pop(0))
        else:
            command = 'KeyToKey'

        keymap = parse_keymap(command, keyargs)
        return (keymap, filters)

    def is_list_or_tuple(self, obj):
        return isinstance(obj, (list, tuple))

    def is_command_format(self, s):
        return s.startswith('(') and s.endswith(')')

    def get_command_value(self, s):
        return s[1:-1]

    def to_xml(self):
        return self.xml_tree


class Block(XML_base):
    def __init__(self, keymaps, filters=None):
        self.keymaps = keymaps
        self.filters = filters or []

    def to_xml(self):
        xml_tree = self.create_tag('block')

        for filter in self.filters:
            xml_tree.append(filter.to_xml())
        for keymap in self.keymaps:
            xml_tree.append(keymap.to_xml())

        return xml_tree


if __name__ == '__main__':
    import doctest
    doctest.testmod(extraglobs={'__version__': __version__})



# -*- coding: utf-8 -*-
import os
import inspect
import subprocess
from easy_karabiner.xml_base import XML_base


class Hashable(object):
    def _id(self):
        raise Exception('Need override')

    def __hash__(self):
        return hash(self._id())

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self._id() == other._id()


def find_all_subclass_of(superclass, global_vars):
    # remove start with '_'
    names = filter(lambda name: not name.startswith('_'), global_vars.keys())
    # remove name which is not a class name
    names = filter(lambda name: inspect.isclass(global_vars[name]), names)
    # remove class name which is not a subclass of superclass
    names = filter(lambda name: issubclass(global_vars[name], superclass), names)
    return list(map(global_vars.get, names))

def remove_all_space(s):
    return ''.join(s.split())

def is_xml_element_equal(node1, node2):
    if len(node1) != len(node2):
        return False
    if node1.tag != node2.tag:
        return False
    if node1.attrib != node2.attrib:
        return False

    text1 = '' if node1.text is None else remove_all_space(node1.text)
    text2 = '' if node2.text is None else remove_all_space(node2.text)
    return text1 == text2

def is_xml_tree_equal(tree1, tree2):
    if is_xml_element_equal(tree1, tree2):
        elems1 = sorted(list(tree1), key=lambda e: (e.tag, e.attrib, e.text))
        elems2 = sorted(list(tree2), key=lambda e: (e.tag, e.attrib, e.text))

        for i in range(len(elems1)):
            if not is_xml_tree_equal(elems1[i], elems2[i]):
                return False
        return True
    else:
        return False

def assert_xml_equal(xml_tree1, xml_tree2):
    if isinstance(xml_tree1, XML_base):
        xml_tree1 = xml_tree1.to_xml()
    else:
        xml_tree1 = XML_base.parse_string(xml_tree1)

    if isinstance(xml_tree2, XML_base):
        xml_tree2 = xml_tree2.to_xml()
    else:
        xml_tree2 = XML_base.parse_string(xml_tree2)

    assert(is_xml_tree_equal(xml_tree1, xml_tree2))

def has_execuable(cmdname):
    with open(os.devnull, "w") as f:
        return subprocess.call(['which', cmdname], stdout=f, stderr=f) == 0

def get_apppath(appname, default=None):
    if not hasattr(get_apppath, '_apps'):
        if has_execuable('mdfind'):
            cmd = ['mdfind', 'kMDItemContentType==com.apple.application-bundle']
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            output = proc.stdout.read().decode('utf-8')

            apppaths = list(filter(lambda s: len(s.strip()) > 0, output.split('\n')))
            appnames = list(map(os.path.basename, apppaths))

            get_apppath._apps = dict(zip(appnames, apppaths))
        else:
            get_apppath._apps = dict()

    return get_apppath._apps.get(appname, default)


def replace_startswith_to(s, startswith, newstart):
    if s.startswith(startswith):
        newval = '%s%s' % (newstart, s.split(startswith, 1)[-1])
        return newval
    else:
        return s


# -*- coding: utf-8 -*-
from __future__ import print_function
from functools import reduce
from itertools import groupby
from easy_karabiner import util
from easy_karabiner.lookup import DefQuery
from easy_karabiner.xml_base import XML_base


class BaseFilter(XML_base, util.Hashable):
    def __init__(self, *vals, **kwargs):
        self.type = kwargs.get('type', 'only')
        self.vals = vals
        self.kwargs = kwargs

    def get_vals(self):
        return self.vals

    def get_tag_name(self):
        header = self.__class__.__name__.lower().rsplit('filter', 1)[0]
        tag_name = '%s_%s' % (header, self.type)
        return tag_name

    def to_xml(self):
        xml_tree = self.create_tag(self.get_tag_name())
        xml_tree.text = ',\n'.join(self.get_vals())
        return xml_tree

    def __add__(self, another):
        if self.get_tag_name() == another.get_tag_name():
            self.vals += another.vals
            return self
        else:
            msg = "Can't add %s with %s" % (self.get_tag_name(), another.get_tag_name())
            raise Exception(msg)

    def _id(self):
        return (self.get_tag_name(), tuple(self.get_vals()))


class Filter(BaseFilter):
    '''
    >>> print(Filter('SKIM'))
    <only>SKIM</only>
    >>> print(Filter('SKIM', type='not'))
    <not>SKIM</not>
    '''

    def get_tag_name(self):
        return self.type


class ReplacementFilter(Filter):
    pass


class DeviceFilter(BaseFilter):
    def get_tag_name(self):
        tag_name = 'device_%s' % self.type
        return tag_name


class DeviceProductFilter(DeviceFilter):
    pass


class DeviceVendorFilter(DeviceFilter):
    pass


class WindowNameFilter(BaseFilter):
    pass


class UIElementRoleFilter(BaseFilter):
    pass


class InputSourceFilter(BaseFilter):
    pass


class ModifierFilter(BaseFilter):
    def get_tag_name(self):
        tag_name = 'modifier_%s' % self.type
        return tag_name


_CLASSES = util.find_all_subclass_of(BaseFilter, globals())

def split_type_and_val(val):
    if val.startswith('!'):
        type = 'not'
        val = val[1:]
    else:
        type = 'only'

    if val.startswith('{{') and val.endswith('}}'):
        val = val[2:-2]

    return (type, val)

# @return [Filter]
def parse_filter(vals):
    type_val_pairs = map(split_type_and_val, vals)

    # create filters
    filters = []
    for type, val in type_val_pairs:
        clsname = DefQuery.query_filter(val) or 'Filter'

        if clsname == 'ReplacementFilter':
            val = '{{%s}}' % val
        elif clsname == 'DeviceProductFilter':
            val = 'DeviceProduct::%s' % val
        elif clsname == 'DeviceVendorFilter':
            val = 'DeviceVendor::%s' % val
        elif clsname == 'ModifierFilter':
            val = 'ModifierFlag::%s' % val

        for cls in _CLASSES:
            if clsname == cls.__name__:
                filter = cls(val, type=type)
                filters.append(filter)
                break

    filters_lists = [list(v) for k, v in groupby(filters, lambda f: f.get_tag_name())]
    filters = list(map(lambda fs: reduce(lambda x, y: x + y, fs), filters_lists))
    return filters


if __name__ == "__main__":
    import doctest
    doctest.testmod()